"""
CLI

Main entrypoint to the program.
"""

import click
from pathlib import Path
from slides_xp import make_app


HELP_TEXT = """
sxp

Slides XP CLI
"""


@click.command("sxp", help=HELP_TEXT, options_metavar="<options>")
@click.argument(
    "paths",
    type=click.Path(
        exists=True,
        dir_okay=True,
        file_okay=False,
        readable=True,
        path_type=Path,
    ),
    nargs=-1,
    required=True,
)
@click.option(
    "--theme",
    envvar="SXP_THEME",
    help="Theme directory to serve CSS stylesheets from",
)
@click.option(
    "-h",
    "--host",
    default="localhost",
    envvar="HOST",
    help="Port to run server on",
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=3000,
    envvar="PORT",
    help="Port to run server on",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Whether to enable Flask debug mode",
)
def cli(
    paths: tuple[Path, ...],
    *,
    theme: str | None,
    host: str,
    port: int,
    debug: bool,
):
    """
    Entrypoint to CLI.
    """
    app = make_app(list(paths), theme)

    app.run(host, port, debug)



from flask import Flask, Blueprint, redirect, send_file, send_from_directory
from pathlib import Path, PosixPath
import pyhtml as p

from slides_xp.picker import Choice, picker
from slides_xp.slide import slide
from slides_xp.util import dir_contains_slides, list_subdirs, slides_list


lib_dir = Path(__file__).parent


def error(code: int, message: str):
    return str(
        p.html(
            p.body(
                p.h1(f"Error {code}"),
                p.p(message),
            )
        )
    ), code


def make_blueprint(name: str, root_dir: Path):
    bp = Blueprint(name, __name__, url_prefix=f"/{name}")

    @bp.get("/", defaults={"path": ""})
    @bp.get("/<path:path>")
    def endpoint(path: str):
        file_path = root_dir / path
        file_url = f"/{name}/{path}".removesuffix("/")
        # Root path must be a parent of `full_path` to prevent escaping the
        # specified directories
        if root_dir not in [file_path, *file_path.parents]:
            return error(403, "Illegal path")
        # 404 if file/dir does not exist
        if not file_path.exists():
            return error(404, "File not found")
        # If file, send it
        if file_path.is_file():
            # Render markdown files as HTML
            if file_path.suffix in [".md", ".py"]:
                return str(slide(file_path, PosixPath(file_url)))
            else:
                return send_file(file_path.absolute())
        elif dir_contains_slides(file_path):
            # Dir with markdown, render a list of slides
            return str(
                picker(
                    str(root_dir),
                    [
                        Choice(p.name, f"{file_url}/{p.name}")
                        for p in slides_list(file_path)
                    ],
                    parent=str(file_path.parent),
                )
            )
        else:
            # Otherwise, dir with no markdown, so render a list of subdirs
            return str(
                picker(
                    str(root_dir),
                    [
                        Choice(p.name, f"{file_url}/{p.name}")
                        for p in list_subdirs(file_path)
                    ],
                    parent=str(file_path.parent),
                )
            )

    return bp


def make_app(paths: list[Path], theme: str | None = None):
    if theme is None:
        theme_dir = lib_dir / "themes" / "default"
    elif Path(theme).is_dir():
        theme_dir = Path(theme)
    elif (lib_dir / "themes" / theme).is_dir():
        theme_dir = lib_dir / "themes" / theme
    else:
        # Invalid theme
        theme_dir = lib_dir / "themes" / "default"

    app = Flask(__name__)

    for path in paths:
        app.register_blueprint(make_blueprint(path.name, path))

    @app.get("/")
    def root():
        """
        Root path.

        If using one path, redirect to it. Otherwise give a choice.
        """
        if len(paths) == 1:
            return redirect(f"/{paths[0].name}/")
        else:
            return str(
                picker(
                    "Slides XP",
                    [Choice(p.name, f"/{p.name}") for p in paths],
                )
            )

    @app.get("/javascript/<path>")
    def scripts(path):
        return send_from_directory(lib_dir / "javascript", path)

    @app.get("/css/<path>")
    def css(path):
        return send_from_directory(lib_dir / "css", path)

    @app.get("/theme/<path>")
    def theme_css(path):
        return send_from_directory(theme_dir, path)

    return app


if __name__ == "__main__":
    app = make_app([Path("temp")])
    app.run(port=3000)



"""
theme.py

Code for registering and interacting with themes.
"""

import pyhtml as p


class Theme:
    def make_slide(self, slide_content: p.Tag) -> p.Tag:
        """
        Given the HTML content for a slide, return the HTML to use for the
        slide.
        """
        return slide_content

    def make_slide_selector(self, slides: list[str]) -> p.Tag:
        """
        Given a list of slides, return HTML for a slide picker.
        """
        return NotImplemented



from pathlib import Path


def list_subdirs(path: Path) -> list[Path]:
    return [subdir for subdir in path.iterdir() if subdir.is_dir()]


def dir_contains_slides(path: Path) -> bool:
    """
    Returns whether the given directory contains Markdown files.
    """
    return next(iter(slides_list(path)), None) is not None


def first_slide(path: Path) -> Path:
    """
    Returns the path of the first slide in the given dir
    """
    return next(iter(slides_list(path)))


def file_is_slide(path: Path):
    return any(path.name.endswith(suffix) for suffix in [".md", ".slide.py"])


def slides_list(path: Path) -> list[Path]:
    """
    Returns a slides list for the given directory.
    """
    return sorted(
        [
            child
            for child in path.iterdir()
            if child.is_file() and file_is_slide(child)
        ],
        key=lambda i: i.name,
    )



"""
Render markdown as HTML
"""

from typing import cast
import mistune
from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import html


# https://mistune.lepture.com/en/latest/guide.html#customize-renderer
class HighlightRenderer(mistune.HTMLRenderer):
    """
    Markdown render that adds code block highlights
    """

    def block_code(self, code, info=None):
        if info:
            lexer = get_lexer_by_name(info, stripall=True)
            formatter = html.HtmlFormatter()
            return highlight(code, lexer, formatter)
        return "<pre><code>" + mistune.escape(code) + "</code></pre>"


markdown = mistune.create_markdown(renderer=HighlightRenderer())


def render_markdown(text: str) -> str:
    """
    Render markdown as HTML.
    """
    # We know it's a string, since we haven't set the renderer to None
    return cast(str, markdown(text))



from dataclasses import dataclass
import pyhtml as p


@dataclass
class Choice:
    name: str
    url: str


def picker(root: str, choices: list[Choice], parent: str | None = None):
    choices_html = [
        p.div(class_="picker-item")(
            p.a(href=f"{choice.url}")(choice.name),
        )
        for choice in choices
    ]

    if parent is not None:
        choices_html.insert(
            0, p.div(class_="picker-item")(p.a(href=parent)(".."))
        )

    return p.html(
        p.head(
            p.title("Slides XP"),
            p.script(src="/javascript/navigator.js", defer=True),
            p.link(href="/css/root.css", rel="stylesheet"),
            p.link(href="/theme/main.css", rel="stylesheet"),
            p.link(href="/theme/picker.css", rel="stylesheet"),
        ),
        p.body(
            p.h1(root),
            p.div(class_="picker-box")(choices_html),
        ),
    )



import importlib
import importlib.util
import sys
import pyhtml as p
from pathlib import Path, PosixPath

from slides_xp.markdown import render_markdown
from slides_xp.util import slides_list


def python_slide(file: Path) -> p.Tag:
    # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    module_name = f"slides_xp.temp.{file.name.removesuffix('.py')}"

    # Add rule file's directory to the module search path, so that imports
    # work as-expected
    sys.path.append(str(file.parent.absolute()))
    # Now begin the import
    spec = importlib.util.spec_from_file_location(module_name, file)
    if spec is None:
        raise ImportError(f"Import spec for file '{file}' was None")

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module

    if spec.loader is None:
        raise ImportError(f"Spec loader for file '{file}' was None")

    # Any exceptions this raises get caught by the calling code
    try:
        spec.loader.exec_module(module)
    except BaseException as e:
        e.add_note(
            f"This exception occurred during execution of file "
            f"{file}. It is unlikely to be an issue with Slides-XP."
        )
        raise e

    try:
        render_fn = module.render
    except AttributeError as e:
        raise AttributeError(
            f"Python slide '{file}' does not have a `render` function"
        ) from e
    if not callable(render_fn):
        raise TypeError("Python slide's `render` attribute is not callable")

    return render_fn()  # type: ignore


def markdown_slide(file: Path) -> p.Tag:
    return p.div(class_="slide-content")(
        p.DangerousRawHtml(render_markdown(file.read_text()))
    )


def slide(file_path: Path, file_url: PosixPath) -> p.html:
    slides = list(slides_list(file_path.parent))

    curr_index = slides.index(file_path)

    first = f"'{file_url.parent / slides[0].name}'" if len(slides) else "null"
    prev = (
        f"'{file_url.parent / slides[curr_index - 1].name}'"
        if curr_index > 0
        else "null"
    )
    next = (
        f"'{file_url.parent / slides[curr_index + 1].name}'"
        if curr_index < len(slides) - 1
        else "null"
    )

    if file_path.suffix == ".py":
        rendered = python_slide(file_path)
    elif file_path.suffix == ".md":
        rendered = markdown_slide(file_path)
    else:
        raise ValueError(f"File '{file_path}' cannot be rendered")

    return p.html(
        p.head(
            p.title("Slides XP"),
            p.script(src="/javascript/navigator.js", defer=True),
            p.link(href="/css/root.css", rel="stylesheet"),
            p.link(href="/theme/main.css", rel="stylesheet"),
            p.link(href="/theme/slide.css", rel="stylesheet"),
        ),
        p.body(
            p.script(f"""
                window.sxp = {{
                    first: {first},
                    prev: {prev},
                    next: {next},
                }};
            """),
            rendered,
        ),
    )



"""
slides_xp / Windows XP theme.
"""

import random
import pyhtml as p


def make_intro_slide(
    author: str,
    title: str,
    copyright: str,
) -> p.div:
    return p.div(class_="bootloader")(
        p.span(class_="bootloader-title")(
            p.h2(author),
            p.sup("®"),
            p.img(class_="windows-logo")(
                alt="Windows XP logo", src="/theme/xp.png"
            ),
            p.sub("TM"),
        ),
        p.span(class_="bootloader-subtitle")(
            p.h1(f"{title}", p.sup(p.small("®"))),
            p.h1(class_="xp")("xp"),
        ),
        p.div(class_="boot-animation")(
            [p.div()] * 3,
        ),
        p.div(class_="gap"),
        p.div(class_="copyright")(f"Copyright © {copyright}"),
    )


def make_bsod_slide(
    slide_type: str,
    instructions: str,
) -> p.div:
    exc_addr = f"{random.randint(0x0000_0000, 0xFFFF_FFFF):08X}"
    vxd_id = f"{random.randint(1, 0xFF):02X}"
    vxd_addr = f"{random.randint(0x1000, 0xFFFF):04X}"
    return p.div(class_="bsod")(
        p.div(class_="bsod-title")("SLIDES XP"),
        p.pre(
            f"A fatal exception has occurred at 00:{exc_addr} in VXD "
            f"sdxp9x({vxd_id}) + 0000{vxd_addr}, The {slide_type} will be "
            "terminated."
        ),
        p.pre(instructions),
        p.span(
            p.pre("Press any key to continue"),
            p.pre(class_="caret")("_"),
        ),
    )
